import omp
import os
import threading
from enum import Enum

import omp.core.threading


class Sched(Enum):
    static = 1
    dynamic = 2
    guided = 3
    auto = 4
    runtime = 5


class InternalControlVariables:

    num_procs_var = os.cpu_count()

    _OMP_NUM_THREADS = 'OMP_NUM_THREADS'
    _nthreads_var = num_procs_var if _OMP_NUM_THREADS not in os.environ else int(os.environ[_OMP_NUM_THREADS])

    @property
    def nthreads_var(self):
        return InternalControlVariables._nthreads_var

    @nthreads_var.setter
    def nthreads_var(self, value):
        InternalControlVariables._nthreads_var = value

    _OMP_NUM_TEAMS = 'OMP_NUM_TEAMS'
    _nteams_var = 0 if _OMP_NUM_TEAMS not in os.environ else int(os.environ[_OMP_NUM_TEAMS])

    @property
    def nteams_var(self):
        return InternalControlVariables._nteams_var

    @nteams_var.setter
    def nteams_var(self, value):
        InternalControlVariables._nteams_var = value

    _OMP_SCHEDULE = 'OMP_SCHEDULE'
    _run_sched_var = (Sched.dynamic, 1) if _OMP_SCHEDULE not in os.environ else (Sched[os.environ[_OMP_SCHEDULE]], 1)

    @property
    def run_sched_var(self):
        return InternalControlVariables._run_sched_var

    @run_sched_var.setter
    def run_sched_var(self, value):
        InternalControlVariables._run_sched_var = value

    def __init__(self, thread: 'omp.core.threading.Thread'):
        self.thread_num_var = thread.rank
        self.team_size_var = thread.team.size


def get_num_procs():
    return threading.current_thread().icv.numprocs_var


def set_num_threads(n: int):
    threading.current_thread().icv.nthreads_var = n


def get_max_threads():
    return threading.current_thread().icv.nthreads_var


def set_num_teams(n: int):
    threading.current_thread().icv.nteams_var = n


def get_max_teams():
    return threading.current_thread().icv.nteams_var


def get_thread_num():
    return threading.current_thread().icv.thread_num_var


def get_num_threads():
    return threading.current_thread().icv.team_size_var


def get_dynamic():
    return False


def set_schedule(kind: Sched, chunk=1):
    threading.current_thread().icv.run_sched_var = (kind, chunk)


def get_schedule():
    return threading.current_thread().icv.run_sched_var



from omp.core.openmp import OpenMP, Clause


@OpenMP.clause('private', ('critical', 'for', 'parallel', 'single'))
class PrivateClause(Clause):

    name = 'private'

    def __init__(self, directive, args):
        super().__init__(directive, args)

        self.directive.privates.update((varname.strip() for varname in args.split(',')))



from omp.core.openmp import OpenMP, Clause


@OpenMP.clause('nowait', ('for', 'single'))
class NoWaitClause(Clause):

    name = 'nowait'

    def __init__(self, directive, args):
        super().__init__(directive, args)

        directive.nowait = True



from omp.core.openmp import Directive, OpenMP
from omp.core.threading import barrier

import ast


@OpenMP.directive('barrier')
class BarrierDirective(Directive):

    """
    OpenMP parallel construct implementation.
    """

    def parse(self, node: ast.With) -> ast.With:
        return node

    def run(self):
        barrier()



import omp.core as core
import omp.directives as directives
import omp.clauses as clauses

from omp.core.openmp import OpenMP
from omp.core.entry import enable

from threading import current_thread

# Avoid linter warnings for package shortcuts definitions.
core
directives
clauses

OpenMP
enable

get_num_procs = core.primitives.get_num_procs
set_num_threads = core.primitives.set_num_threads
get_max_threads = core.primitives.get_max_threads
set_num_teams = core.primitives.set_num_teams
get_max_teams = core.primitives.get_max_teams
get_thread_num = core.primitives.get_thread_num
get_num_threads = core.primitives.get_num_threads
get_dynamic = core.primitives.get_dynamic
set_schedule = core.primitives.set_schedule
get_schedule = core.primitives.get_schedule

current_thread().icv = core.primitives.InternalControlVariables(current_thread())



import omp.core.openmp
from omp.core.ast_tools import LinenoStripper

import ast
import copy
import inspect
import textwrap
import threading

from types import CodeType, FunctionType


class OpenMPTransformer(ast.NodeTransformer):
    """
    Recursively find the OpenMP constructs and replace them with their implementations.
    """

    def __init__(self, locs=None, globs=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.locs = locs
        self.globs = globs

    def visit_With(self, node: ast.With) -> ast.With:
        # We need to make sure that this is an OpenMP construct.

        # The with statement should use only one context manager.
        if len(node.items) != 1:
            return self.generic_visit(node)

        # Bulletproofing.
        if not isinstance(node.items[0], ast.withitem):
            return self.generic_visit(node)

        # We are now sure we have a withitem.
        item: ast.withitem = node.items[0]

        # The context manager should be an OpenMP instanciation.
        # This means calling the constructor.

        if not isinstance(item.context_expr, ast.Call):
            return self.generic_visit(node)

        call: ast.Call = item.context_expr

        # Check that the function is referenced by Attributes of a Name, or a Name directly.
        ref = call.func
        while isinstance(ref, ast.Attribute):
            ref: ast.Attribute
            # Bulletproofing
            if not isinstance(ref.ctx, ast.Load):
                return self.generic_visit(node)
            ref = ref.value

        if not isinstance(ref, ast.Name):
            return self.generic_visit(node)

        name: ast.Name = ref

        # Bulletproofing.

        if not isinstance(name.ctx, ast.Load):
            return self.generic_visit(node)

        # In order to check that this call indeed an OpenMP instanciation,
        # we will evaluate the name being called in the function's definition namespace.

        expr: ast.Expression = ast.Expression(call.func)
        try:
            called = eval(compile(expr, filename='<OMP Parser>', mode='eval'), self.globs, self.locs)
        except NameError:
            # The name we are trying to evaluate could be undefined in the context.
            return self.generic_visit(node)

        if called is not omp.core.openmp.OpenMP:
            return self.generic_visit(node)

        # We are now sure this is an OpenMP construct. (Not necessarily a valid one.)
        # We run the found instanciation and run its logic on the found construct.
        instruction = eval(compile(ast.Expression(call), filename='<OMP Parser>', mode='eval'), self.globs, self.locs)

        # Parsing this node **after** its children allows us to know the exhaustive list of local variables that will be
        # involved in the inner function definitions.
        return instruction._parse_With(ast.fix_missing_locations(self.generic_visit(node)))


class EnableFunction(ast.NodeTransformer):
    """
    Transforms a enable-decorated function definition into an enabled function
    definition without the enable decorator.
    """

    def __init__(self, locs=None, globs=None, varnames=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.locs = locs
        self.globs = globs
        self.varnames = varnames
        if self.varnames is None:
            self.varnames = []

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        new = copy.deepcopy(node)
        # Remove the last decorator which should be ours if the user followed our documentation.
        new.decorator_list = new.decorator_list[:-1]
        # Inject a known name in the namespace for our library.
        new.body = [LinenoStripper().visit(ast.parse('import omp as _omp_internal', mode='exec').body[0])] + new.body

        # Run OpenMPTransformer until the result stabilizes.
        old_ver: str = ''
        while old_ver != ast.dump(new):
            old_ver = ast.dump(new)
            new = OpenMPTransformer(self.globs, self.locs).visit(ast.fix_missing_locations(new))

        return new


def enable(*args, **kwargs):
    """
    Enable OpenMP in the given block of code.

    Note: If your function has several decorators, this one should be the first to run. (i.e. closest to the function definition)

    Use as a decorator:
    ```
    @omp.enable
    def main():
        <omp-enabled code...>
    ```

    # Call your function
    main()
    """

    def decorator(function):
        threading.current_thread().omp_parsing = True
        # The user code calls `enable` which itself calls `decorator`.
        # We need to go back two stack frames.
        caller_frame = inspect.currentframe().f_back.f_back
        globs, locs = caller_frame.f_globals, caller_frame.f_locals

        # We need _omp_internal for our own compilations.
        locs.update({'_omp_internal': omp})

        # Retrieve the source code of the decorated function.
        src: str = textwrap.dedent(inspect.getsource(function))

        # Convert the source code to ast.
        src_ast: ast.Module = ast.parse(src, mode='exec')
        src_ast = ast.increment_lineno(src_ast, caller_frame.f_lineno - src_ast.body[0].lineno + 1)

        # Patch the source ast.
        # We need to make sure that each node has a line number.
        # Since the initial function was already compiled a first time, we can recover the
        # local variables the function uses from its code object.
        patched_ast = ast.fix_missing_locations(
            EnableFunction(globs, locs, function.__code__.co_varnames).visit(src_ast)
            )

        # ALERT: Remove this debug print. (Shows the final transformed source code.)
        # print(ast.unparse(patched_ast))

        # Compile the patched ast.
        patched: CodeType = compile(patched_ast, filename=inspect.getsourcefile(function), mode='exec')

        # redefine the function in the initial context.
        # TODO: Allow enabling a function with closure (nested enabled function)
        #       Note: This would involve locating the function definition in the module's source code and recompile the outer function entirely.
        exec(patched, globs, locs)

        threading.current_thread().omp_parsing = False

        return locs[function.__name__]

    # Simple decorator.
    if len(args) == 1 and isinstance(args[0], FunctionType):
        return decorator(args[0])

    # TODO: Parametrized decorator support.
    # TODO: If statement support.
    return decorator



from omp.core.openmp import OpenMP, Clause
from omp.core.primitives import Sched

import omp

@OpenMP.clause('schedule', ('for',))
class ScheduleClause(Clause):

    name = 'schedule'

    def __init__(self, directive, args):
        super().__init__(directive, args)

        split_args = args.split(',')

        directive.schedule = (Sched[split_args[0]], int(split_args[-1]) if len(split_args) > 1 else omp.get_schedule()[1])



from omp.core.openmp import OpenMP, Clause


@OpenMP.clause('reduction', ('for'))
class ReductionClause(Clause):

    name = 'reduction'

    def __init__(self, directive, args):
        super().__init__(directive, args)

        operator, varname = args.split(':')

        self.directive.privates.add(varname.strip())
        self.directive.reduction.update({varname: operator})


operators = {
    '+': lambda a, b: a + b,
    '*': lambda a, b: a * b,
    '&': lambda a, b: a & b,
    '|': lambda a, b: a | b,
    '^': lambda a, b: a ^ b,
    '&&': lambda a, b: a and b,
    '||': lambda a, b: a or b,
    None: lambda a, b: a
}



import ast
import threading


class Clause:

    def __init__(self, directive: 'Directive', args: str):
        self.directive = directive
        self.args = args

        self.directive.clauses.append(self)


class Directive:

    def __init__(self, openMP: 'OpenMP'):
        self.openMP = openMP
        self.clauses: list(Clause) = []

        self.privates = set()
        self.nowait = False
        self.reduction = dict()

    def parse(self, node: ast.With) -> ast.With:
        """
        Convert the given With OpenMP construct with its implementation.
        """
        # raise NotImplementedError()
        pass

    def run(self):
        """
        Implementation of an OpenMP directive
        """
        # raise NotImplementedError()
        pass

    @staticmethod
    def replace(target: list[ast.AST], content: list[ast.AST], match: ast.AST = ast.Pass) -> list[ast.AST]:
        """
        Return a new list where the elements of content are where there is a Pass in target.
        """
        res = []
        for el in target:
            if isinstance(el, match):
                res.extend(content)
            else:
                res.append(el)
        return res

    @staticmethod
    def assign_shared(shared: list[str], value: ast.AST = None) -> ast.Assign:
        """
        Return a tuple assignment to the shared variables listed.
        """

        # Apparently, assigning an empty list to an empty tuple is valid python syntax.
        return [ast.Assign(
            targets=[ast.Tuple(elts=[ast.Name(id=name, ctx=ast.Store()) for name in shared],
                               ctx=ast.Store())],
            value=ast.BinOp(
                left=ast.List(elts=[ast.Constant(value=value)], ctx=ast.Load()),
                op=ast.Mult(),
                right=ast.Constant(value=len(shared))
                )
            )]

    # @staticmethod
    def list_locals(self, body: list[ast.AST]) -> list[str]:
        """
        Return a list of the local variables used in the given function body.
        """

        template = """\
def _omp_internal_inner_func():
    pass
"""

        ast_template = ast.parse(template, mode='exec')

        inner_func = ast_template.body[0]

        inner_func.body = Directive.replace(inner_func.body, body)

        # To find the list of local variables defined in the function, we need to compile it.
        globs, locs = dict(), dict()
        exec(compile(ast_template, '<OMP Parser>', mode='exec'), globs, locs)

        return [varname for varname in set(locs['_omp_internal_inner_func'].__code__.co_varnames + locs['_omp_internal_inner_func'].__code__.co_cellvars) - self.privates if not varname.startswith('_omp_internal')]


class OpenMP:

    """
    This class represents a call to an OpenMP directive.

    When used in omp-enabled user code, this class calls an OpenMP directive.

    When this class is used as a context manager, the with statement is replaced
    by the implementation of the construct.
    """

    # All the directives supported by the library. This variable is global.
    # To register a new directive, update omp.openmp.OpenMP.directives.
    directives: dict[str, Directive] = {}
    clauses: dict[str, Clause] = {}

    dir_impl: Directive = None

    @staticmethod
    def directive(name: str):
        """
        This decorator is meant for internal use.

        Register a new directive.
        """

        def decorator(cls):
            OpenMP.directives.update({name: cls})
        return decorator

    @staticmethod
    def clause(name: str, directives: list[str]):
        """
        This decorator is meant for internal use.

        Register a new clause.
        """
        def decorator(cls):
            OpenMP.clauses.update({name: (cls, directives)})
        return decorator

    def parse_clauses(self, clauses, directive):
        i = 0
        while i < len(clauses):
            if clauses[i] in ('(', ',', ' '):
                break
            i += 1

        clause = clauses[:i]
        args = ''

        if i < len(clauses):
            j = i + 1
            if clauses[i] == '(':
                while j < len(clauses):
                    if clauses[j] == ')':
                        break
                    j += 1
            args = clauses[i+1:j]
            self.parse_clauses(clauses[j+1:].strip(',').strip(), directive)

        if clause in self.clauses:
            cls, dirs = self.clauses[clause]
            if directive in dirs:
                cls(self.dir_impl, args)

    def __init__(self, arg: str = ''):
        words: list(str) = arg.split()

        directive: str = words[0] if words else arg

        # Allow for `parallel for` and future shortcut directives.
        for word in words[1:]:
            if f'{directive} {word}' not in self.directives:
                break
            directive = f'{directive} {word}'

        if directive in OpenMP.directives:
            self.dir_impl = OpenMP.directives[directive](self)
            self.clause_str = arg[len(directive):].strip()
            self.parse_clauses(arg[len(directive):].strip(), directive)

        if not threading.current_thread().omp_parsing and self.dir_impl is not None:
            self.dir_impl.run()

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        return False

    def _parse_With(self, node: ast.With) -> ast.With:
        """
        This method is meant for internal use.

        Replace this OpenMP context manager by its implementation.
        """

        if self.dir_impl is not None:
            return self.dir_impl.parse(node)

        return node



from omp.core.openmp import Directive, OpenMP
from omp.core.ast_tools import LinenoStripper

import ast


@OpenMP.directive('parallel for')
class ParallelForConstruct(Directive):

    """
    OpenMP parallel construct implementation.
    """

    @property
    def template(self):
        return f"""\
with _omp_internal.core.openmp.OpenMP("parallel {self.openMP.clause_str}"):
    with _omp_internal.core.openmp.OpenMP("for {self.openMP.clause_str}"):
        pass # Replaced by user code
        """

    def parse(self, node: ast.With) -> ast.With:
        # Parse the template to AST.
        ast_template = LinenoStripper().visit(ast.parse(self.template, mode='exec'))

        # Extract the inner with statement.
        with_stmt: ast.With = ast_template.body[0].body[0]

        # Replace the pass statement in the inner with body.
        with_stmt.body = self.replace(with_stmt.body, node.body)
        return ast_template.body[0]



from omp.core.openmp import Directive, OpenMP
from omp.core.threading import Team
from omp.core.ast_tools import LinenoStripper

import ast
import random


@OpenMP.directive('parallel')
class ParallelConstruct(Directive):

    """
    OpenMP parallel construct implementation.
    """

    @property
    def template(self):
        nonce = random.randint(0, 100000)
        return f"""\
with _omp_internal.core.openmp.OpenMP():
    if False:
        pass # Replaced by shared variables declarations
    @_omp_internal.directives.parallel_construct.run_parallel
    def _omp_internal_inner_func{nonce}():
        pass # Replaced by user code
    _omp_internal_inner_func{nonce}()
        """

    def parse(self, node: ast.With) -> ast.With:
        # Parse the template to AST.
        ast_template = LinenoStripper().visit(ast.parse(self.template, mode='exec'))

        # Extract the if statement.
        if_stmt: ast.If = ast_template.body[0].body[0]

        # Extract the inner function definition.
        inner_func: ast.FunctionDef = ast_template.body[0].body[1]

        shared = self.list_locals(node.body)

        # Replace the pass statement in the if body.
        if_stmt.body = self.replace(if_stmt.body, self.assign_shared(shared))

        # List the shared variables
        nonlocals = []
        if len(shared) > 0:
            nonlocals = [ast.Nonlocal(names=list(shared))]

        # Replace the pass statement in the inner function body.
        inner_func.body = self.replace(inner_func.body, nonlocals + node.body)
        return ast_template.body[0]


def run_parallel(func):
    """
    When the new function is called, creates a team of threads that will each run the given function concurrently.
    Decorates the given function.
    """

    def wrapped(*args, **kwargs):
        team = Team(size=None, target=func, args=args, kwargs=kwargs)

        team.start()
        team.join()
    return wrapped



import threading

import omp.core.primitives
import omp


class Thread(threading.Thread):
    """
    Represents a thread managed by omp
    """

    def __init__(self, rank=None, team=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rank = rank
        self.team: Team = team
        self.icv = omp.core.primitives.InternalControlVariables(self)
        self.omp_parsing = False


class Team:

    """
    Represents a team of threads.
    """

    OMP_NUM_THREADS = 'OMP_NUM_THREADS'

    def __init__(self, size=None, *args, **kwargs):

        # The default team size is set by the OMP_NUM_THREADS environment variable.
        # If OMP_NUM_THREADS is not set, we use the number of available CPUs as the default size.
        if size is None:
            size = omp.get_max_threads()

        self.size = size
        self.threads = [Thread(i, self, *args, **kwargs) for i in range(self.size)]

        if size > 0:
            self.barrier = threading.Barrier(size)
        self.lock = threading.Lock()

        # ALERT: Access atomically.
        self.singleThread = None

        self.globalvars = {}

    def start(self):
        for thread in self.threads:
            thread.start()

    def join(self):
        for thread in self.threads:
            thread.join()


def barrier():
    threading.current_thread().team.barrier.wait()


# Add our attributes to the main thread.
_mainThread = threading.current_thread()
_mainThread.rank = 0
_mainThread.team = Team(size=0)
_mainThread.team.size = 1
_mainThread.barrier = threading.Barrier(1)
_mainThread.team.threads.append(_mainThread)
_mainThread.omp_parsing = False



import ast
import copy


class LinenoStripper(ast.NodeTransformer):
    """
    Removes line number information from all nodes.

    This is usefull to ensure that templates don't get wrong line numbers attributed.'
    """

    def visit(self, node: ast.AST):
        new: ast.AST = copy.copy(node)

        for attr in ('lineno', 'col_offset', 'end_lineno', 'end_col_offset'):
            if hasattr(new, attr):
                delattr(new, attr)

        return self.generic_visit(new)



import omp
from omp.core.openmp import Directive, OpenMP
from omp.core.ast_tools import LinenoStripper
from omp.core.primitives import Sched

import ast
import random
import time
import threading
import itertools


def generator_static(it, nonce, chunk):
    """
    When called within a thread of a team, yields the iterations for the current thread.

    Overall, when all the threads of the team call this generator, all the elements of the iterator are yielded.
    """
    for i, el in enumerate(it):
        if i % omp.get_num_threads() == omp.get_thread_num():
            yield el


class EndOfQueue:
    pass


@OpenMP.directive('for')
class ForConstruct(Directive):

    schedule = (Sched.runtime, 1)

    """
    OpenMP for construct implementation.
    """

    @property
    def template(self):

        reduction_vars, reduction_operators = list(zip(*self.reduction.items())) if self.reduction else ([], [])
        nonce = random.randint(0, 100000)

        return f"""\
with _omp_internal.core.openmp.OpenMP():
    if False:
        pass # Replaced by shared variables declarations
    def _omp_internal_inner_func{nonce}({','.join(reduction_vars)}):
        pass # Replaced by user code
        {'#' if self.nowait else ''}_omp_internal.core.openmp.OpenMP("barrier")
        {'' if reduction_vars else '#'}return ({','.join(reduction_vars)},)
    def _omp_internal_inner_func_protect{nonce}():
        {'' if reduction_vars else '#'}nonlocal {','.join(reduction_vars)}
        _omp_internal_retval = _omp_internal_inner_func{nonce}({','.join(reduction_vars)})
        {'' if reduction_vars else '#'}with _omp_internal.core.openmp.OpenMP('critical'):
            {'' if reduction_vars else '#'}({','.join(reduction_vars)},) = [_omp_internal.clauses.reduction.operators[_omp_internal_operator](_omp_internal_var, _omp_internal_val) for _omp_internal_operator, _omp_internal_var, _omp_internal_val in zip({reduction_operators}, ({','.join(reduction_vars)},), _omp_internal_retval)]
    _omp_internal_inner_func_protect{nonce}()
    """

    @staticmethod
    def define_generators():

        if hasattr(omp.directives.for_construct, 'dynamic'):
            return

        @omp.enable
        def generator_dynamic(it, nonce, chunk):
            thread = threading.current_thread()
            with OpenMP("single"):
                thread.team.globalvars[f'_omp_interal_for_batched_it{nonce}'] = iter(itertools.batched(it, chunk)), threading.Lock()
            batched_it, lock = thread.team.globalvars[f'_omp_interal_for_batched_it{nonce}']
            while True:
                try:
                    with lock:
                        el = next(batched_it)
                except StopIteration:
                    break
                else:
                    yield from el
            OpenMP("barrier")
            with OpenMP("single nowait"):
                del thread.team.globalvars[f'_omp_interal_for_batched_it{nonce}']

        omp.directives.for_construct.generator_dynamic = generator_dynamic
        omp.directives.for_construct.generator_guided = omp.directives.for_construct.generator_dynamic
        omp.directives.for_construct.generator_auto = omp.directives.for_construct.generator_dynamic

    def parse(self, node: ast.With):

        self.define_generators()

        for_node: ast.For = node.body[0]

        schedule = self.schedule

        if schedule[0] == Sched.runtime:
            schedule = omp.get_schedule()

        # Wrap the loop iterator in our thread-distributing generator.
        for_node.iter = ast.Call(LinenoStripper().visit(ast.parse(f'_omp_internal.directives.for_construct.generator_{schedule[0].name}')).body[0].value, args=[for_node.iter, ast.Constant(value=time.time()), ast.Constant(value=schedule[1])], keywords=[])

        # We need to protect the target.
        # ALERT: We need to handle unpacking as well. (`for i,j in it`)
        target = for_node.target.id

        # Parse the template to AST.
        ast_template = LinenoStripper().visit(ast.parse(self.template, mode='exec'))

        # Extract the if statement.
        if_stmt: ast.If = ast_template.body[0].body[0]

        # Extract the inner function definition.
        inner_func: ast.FunctionDef = ast_template.body[0].body[1]

        # Variables are shared except for target
        shared = [name for name in self.list_locals(for_node.body) if name != target]

        # Replace the pass statement in the if body.
        if_stmt.body = self.replace(if_stmt.body, self.assign_shared(shared + list(self.reduction.keys())))

        # Replace the pass statement in the inner function body.
        nonlocals = []
        if len(shared) > 0:
            nonlocals = [ast.Nonlocal(names=[name for name in shared if name != target])]

        inner_func.body = self.replace(inner_func.body, nonlocals + node.body)
        return ast_template.body[0]



import omp.directives.parallel_construct as parallel_construct
import omp.directives.for_construct as for_construct
import omp.directives.parallel_for_construct as parallel_for_construct
import omp.directives.single_construct as single_construct
import omp.directives.barrier_directive as barrier_directive
import omp.directives.critical_construct as critical_construct

# Avoid linter warnings for package shortcuts definitions.
parallel_construct
for_construct
parallel_for_construct
single_construct
barrier_directive
critical_construct



import omp.core.entry as entry
import omp.core.openmp as openmp
import omp.core.threading as threading
import omp.core.ast_tools as ast_tools

# Avoid linter warnings for package shortcuts definitions.
entry
openmp
threading
ast_tools
